/*
 * fbwhiptail.c : Framebuffer Cairo Menu
 *
 * Copyright (C) Youness Alaoui (KaKaRoTo)
 * Written in 2011 for PS3, rewritten in 2018
 *
 * This software is distributed under the terms of the GNU General Public
 * License ("GPL") version 3, as published by the Free Software Foundation.
 *
 *
*/


#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <time.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>

#include <sys/time.h>

#include "fbwhiptail_menu.h"

#define TEXT_PAD 5

cairo_surface_t *
create_gradient_background (int width, int height,
    float start_r, float start_g, float start_b,
    float end_r, float end_g, float end_b)
{
  cairo_surface_t * background = NULL;
  cairo_pattern_t *linpat = NULL;
  cairo_t *grad_cr = NULL;

  background = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, width, height);

  linpat = cairo_pattern_create_linear (0, 0, width, height);
  cairo_pattern_add_color_stop_rgb (linpat, 0, start_r, start_g, start_b);
  cairo_pattern_add_color_stop_rgb (linpat, 1, end_r, end_g, end_b);

  grad_cr = cairo_create (background);
  cairo_set_source (grad_cr, linpat);
  cairo_paint (grad_cr);
  cairo_destroy (grad_cr);
  cairo_pattern_destroy (linpat);
  cairo_surface_flush (background);

  return background;
}

cairo_surface_t *
load_image_and_scale (char *path, int width, int height)
{
  cairo_surface_t *image = NULL;
  cairo_surface_t *surface = NULL;

  image = cairo_image_surface_create_from_png (path);
  if (image) {
    cairo_t *cr;
    int img_width, img_height;

    surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32,
        width, height);
    cr = cairo_create (surface);

    cairo_utils_get_surface_size (image, &img_width, &img_height);

    cairo_scale (cr, (float) width / img_width, (float) height / img_height);
    cairo_set_source_surface (cr, image, 0, 0);

    /* Avoid getting the edge blended with 0 alpha */
    cairo_pattern_set_extend (cairo_get_source(cr), CAIRO_EXTEND_PAD);

    /* Replace the destination with the source instead of overlaying */
    cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
    cairo_paint (cr);
    cairo_destroy (cr);
  }

  return surface;
}


void
draw_background (Menu *menu, cairo_t *cr)
{
  if (menu->background) {
    cairo_set_source_surface (cr, menu->background, 0, 0);
    cairo_paint (cr);
  }
}

static void
create_standard_menu_frame (Menu *menu)
{
  cairo_font_extents_t fex;
  cairo_text_extents_t tex;
  cairo_pattern_t *linpat = NULL;
  cairo_hires_buffer_t *frame = NULL;
  int width, height;
  cairo_t *cr;
  int x, y;

  /* Adapt frame height depending on items in the menu */
  width = STANDARD_MENU_FRAME_WIDTH;
  height = cairo_menu_item_rows(menu->menu) * STANDARD_MENU_ITEM_TOTAL_HEIGHT;
  height += cairo_hires_buffer_logical_height (menu->text.surface);
  if (height > STANDARD_MENU_HEIGHT)
    height = STANDARD_MENU_HEIGHT;
  height += STANDARD_MENU_FRAME_HEIGHT;

  frame = cairo_hires_buffer_create (CAIRO_FORMAT_ARGB32, width, height,
    cairo_hires_buffer_scale (cairo_menu_get_hires_buffer (menu->menu)));

  // Draw silver border around the frame
  cr = cairo_hires_buffer_cairo (frame);
  cairo_utils_clip_round_edge (cr, width, height,
      STANDARD_MENU_FRAME_CORNER_RADIUS, STANDARD_MENU_FRAME_CORNER_RADIUS,
      STANDARD_MENU_FRAME_CORNER_RADIUS);
  linpat = cairo_pattern_create_linear (width, 0, width, height);

  cairo_pattern_add_color_stop_rgb (linpat, 0.0, 0.3, 0.3, 0.3);
  cairo_pattern_add_color_stop_rgb (linpat, 1.0, 0.8, 0.8, 0.8);

  cairo_set_source (cr, linpat);
  cairo_paint (cr);
  cairo_pattern_destroy (linpat);

  // Create frame gradient
  linpat = cairo_pattern_create_linear (0, 0, width, height);

  cairo_pattern_add_color_stop_rgb (linpat, 0.0, 0.5, 0.5, 0.5);
  cairo_pattern_add_color_stop_rgb (linpat, 1.0, 0.1, 0.1, 0.1);

  cairo_utils_clip_round_edge (cr, width, height,
      STANDARD_MENU_FRAME_CORNER_RADIUS + STANDARD_MENU_FRAME_BORDER_WIDTH,
      STANDARD_MENU_FRAME_CORNER_RADIUS + STANDARD_MENU_FRAME_BORDER_WIDTH,
      STANDARD_MENU_FRAME_CORNER_RADIUS);

  cairo_set_source (cr, linpat);
  cairo_paint (cr);
  cairo_pattern_destroy (linpat);

  cairo_select_font_face(cr, "Arial",
      CAIRO_FONT_SLANT_NORMAL,
      CAIRO_FONT_WEIGHT_BOLD);

  cairo_set_font_size(cr, STANDARD_MENU_TITLE_FONT_SIZE);

  cairo_font_extents (cr, &fex);
  cairo_text_extents (cr, menu->title, &tex);

  y = (STANDARD_MENU_FRAME_TOP / 2) + (fex.ascent / 2);
  x = ((width - tex.width) / 2) - tex.x_bearing;
  cairo_move_to(cr, x, y);
  cairo_set_source_rgb (cr, 0.5, 0.5, 0.5);
  cairo_show_text (cr, menu->title);

  /* Create the frame with a dropshadow */
  menu->frame = cairo_utils_surface_add_dropshadow (frame,
      FRAME_DROPSHADOW_DISTANCE);
  cairo_hires_buffer_destroy (frame);
}

static void
refresh_text_surface (Menu *menu)
{
  double x;
  double y;
  double width = menu->width * 0.8;
  double height = menu->height * 0.8;
  char **line;
  int cnt = menu->text.start_line;
  cairo_t *cr;

  cr = cairo_hires_buffer_cairo (menu->text.surface);
  x = 0;
  y = 0;

  line = menu->text.lines;
  while (*line != NULL && y + menu->text_size < height) {
    if (cnt > 0) {
      cnt--;
      line++;
      continue;
    }

    /* Drawing text line */
    cairo_save (cr);

    cairo_select_font_face (cr,
        "monospace",
        CAIRO_FONT_SLANT_NORMAL,
        CAIRO_FONT_WEIGHT_BOLD);

    cairo_set_font_size (cr, menu->text_size);

    cairo_set_source_rgb (cr, 1, 1, 1);
    cairo_move_to (cr, x, y + menu->text_size + TEXT_PAD);
    cairo_show_text (cr, *line);
    cairo_restore (cr);
    y += menu->text_size;
    line++;
  }

}

static void
draw_standard_menu (Menu *menu, cairo_t *cr)
{
  int w, h;
  int menu_width;
  int menu_height;
  int text_height;

  if (menu->frame == NULL) {
    create_standard_menu_frame (menu);
  }

  w = cairo_hires_buffer_logical_width (menu->frame);
  h = cairo_hires_buffer_logical_height (menu->frame);
  text_height = cairo_hires_buffer_logical_height (menu->text.surface);

  cairo_menu_redraw (menu->menu);

  cairo_hires_buffer_paint (menu->frame, cr, (menu->width - w) / 2,
      (menu->height - h) / 2);

  cairo_hires_buffer_paint (menu->text.surface, cr,
      (menu->width - w) / 2 + STANDARD_MENU_FRAME_SIDE,
      (menu->height - h) / 2 + STANDARD_MENU_FRAME_TOP);

  menu_width = STANDARD_MENU_WIDTH;
  menu_height = cairo_menu_item_rows(menu->menu) * STANDARD_MENU_ITEM_TOTAL_HEIGHT;

  if (menu_height > STANDARD_MENU_HEIGHT - text_height)
    menu_height = STANDARD_MENU_HEIGHT - text_height;

  cairo_save (cr);
  cairo_translate (cr, ((menu->width - w) / 2) + STANDARD_MENU_FRAME_SIDE,
      ((menu->height - h) / 2) + STANDARD_MENU_FRAME_TOP + text_height);

  if (!menu->gauge) {
    /* Draw a frame around the menu so cut off buttons don't appear clippped */
    cairo_utils_clip_round_edge (cr, menu_width, menu_height, 20, 20, 20);
    cairo_set_source_rgb (cr, 0, 0, 0);
    cairo_paint_with_alpha (cr, 0.5);
  }

  cairo_hires_buffer_paint (cairo_menu_get_hires_buffer (menu->menu), cr, 0, 0);
  cairo_restore (cr);
}

static cairo_surface_t *
create_standard_background (int width, int height, double display_scale,
  float r, float g, float b) {
  cairo_hires_buffer_t *bg;
  cairo_hires_buffer_t *background;
  cairo_surface_t *bg_surface;
  cairo_t *cr;
  cairo_pattern_t *linpat = NULL;

  background = cairo_hires_buffer_create (CAIRO_FORMAT_ARGB32, width, height,
    display_scale);

  bg = cairo_menu_create_hires_default_background (width - (2 * STANDARD_MENU_BOX_X),
      height - (2 * STANDARD_MENU_BOX_Y), display_scale, r, g, b);

  // Draw the grey/silver gradient (for the border)
  cr = cairo_hires_buffer_cairo (background);
  cairo_utils_clip_round_edge (cr, width, height,
      STANDARD_MENU_BOX_CORNER_RADIUS, STANDARD_MENU_BOX_CORNER_RADIUS,
      STANDARD_MENU_BOX_CORNER_RADIUS);

  linpat = cairo_pattern_create_linear (width, 0, width, height);

  cairo_pattern_add_color_stop_rgb (linpat, 0.0, 0.3, 0.3, 0.3);
  cairo_pattern_add_color_stop_rgb (linpat, 1.0, 0.7, 0.7, 0.7);

  cairo_set_source (cr, linpat);
  cairo_paint_with_alpha (cr, 0.5);

  // Clear the inside of the gradient in order to create a border
  cairo_utils_clip_round_edge (cr, width, height,
      STANDARD_MENU_BOX_CORNER_RADIUS + STANDARD_MENU_BOX_BORDER_WIDTH,
      STANDARD_MENU_BOX_CORNER_RADIUS + STANDARD_MENU_BOX_BORDER_WIDTH,
      STANDARD_MENU_BOX_CORNER_RADIUS);

  cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
  cairo_paint (cr);

  // Add a 40% alpha black filter inside, creating a frame around the button
  cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
  cairo_set_source_rgb (cr, 0, 0, 0);
  cairo_paint_with_alpha (cr, 0.4);

  // Draw the actual button
  cairo_hires_buffer_paint (bg, cairo_hires_buffer_cairo(background),
    STANDARD_MENU_BOX_X, STANDARD_MENU_BOX_Y);
  cairo_pattern_destroy (linpat);
  cairo_hires_buffer_destroy (bg);

  bg_surface = cairo_surface_reference(cairo_hires_buffer_surface(background));
  cairo_hires_buffer_destroy(background);

  return bg_surface;
}

char *
load_text_from_file (char *filename)
{
 FILE* fd = NULL;
 char *text = NULL;
 int filesize;

 fd = fopen(filename,"rb");
 if (fd != NULL) {
   fseek(fd, 0, SEEK_END);
   filesize = ftell(fd);
   fseek(fd, 0, SEEK_SET);

   text = (char*) malloc(filesize + 1);
   fread(text, 1, filesize, fd);
   text[filesize] = 0;
   fclose(fd);
 }

 return text;
}

void
create_text_surface (Menu *menu, char *text, int text_size, double display_scale)
{
  char *ptr, *ptr2;
  int filesize;
  int lines;

 memset (&menu->text, 0, sizeof(MenuText));

 /* Calculate number of lines */
 lines = 0;
 ptr = ptr2 = text;
 while (*ptr != 0) {
   if (*ptr == '\\' && ptr[1] == 'n') {
     *ptr2 = '\n';
     ptr += 2;
   } else {
     *ptr2 = *ptr;
     ptr++;
   }
   if (*ptr2++ == '\n')
     lines++;
 }
 *ptr2 = '\0';
 /* Last line */
 lines++;

 menu->text.lines = malloc ((lines + 1) * sizeof(char *));
 menu->text.nlines = 0;
 ptr = text;
 while (*ptr != 0) {
   if (*ptr != '\r')
     menu->text.lines[menu->text.nlines++] = ptr;

   while (*ptr != 0 && *ptr != '\r' && *ptr != '\n')
     ptr++;
   if (*ptr == '\r')
     *ptr++ = 0;
   if (*ptr == '\n')
     *ptr++ = 0;
 }
 menu->text.lines[menu->text.nlines] = NULL;
 menu->text_size = text_size;

 menu->text.surface = cairo_hires_buffer_create (CAIRO_FORMAT_ARGB32,
     STANDARD_MENU_WIDTH, 10 + (text_size + TEXT_PAD) * menu->text.nlines,
     display_scale);
}

Menu *
standard_menu_create (const char *title, char * text, int text_size,
    int width, int height, double display_scale, int rows, int columns)
{
  cairo_surface_t *background, *selected_background, *disabled;
  cairo_t *cr;
  Menu * menu = malloc (sizeof(Menu));
  int text_height;
  int button_width, button_height;

  button_width = STANDARD_MENU_ITEM_BOX_WIDTH / columns - STANDARD_MENU_PAD_X * (columns-1);
  button_height = STANDARD_MENU_ITEM_BOX_HEIGHT;

  memset (menu, 0, sizeof(Menu));
  menu->draw = draw_standard_menu;
  menu->title = title;
  menu->width = width;
  menu->height = height;

  create_text_surface (menu, text, text_size, display_scale);
  refresh_text_surface (menu);
  text_height = cairo_hires_buffer_logical_height (menu->text.surface);
  background = create_standard_background (button_width, button_height,
    display_scale, 0, 0, 0);
  selected_background = create_standard_background (button_width, button_height,
    display_scale, 0.4, 0.4, 0.4);
  disabled = cairo_image_surface_create (CAIRO_FORMAT_ARGB32,
      button_width, button_height);

  cr = cairo_create (disabled);

  cairo_set_source_rgba (cr, 0.7, 0.7, 0.7, 0.7);
  cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
  cairo_utils_clip_round_edge (cr, button_width, button_height,
      STANDARD_MENU_BOX_X + 7, STANDARD_MENU_BOX_Y + 7, 7);
  cairo_paint (cr);

  cairo_destroy (cr);
  cairo_surface_flush (disabled);

  /* Infinite vertical scrollable menu */
  menu->menu = cairo_menu_new_full (STANDARD_MENU_WIDTH,
      STANDARD_MENU_HEIGHT - text_height, display_scale, rows, columns,
      button_width, button_height,
      STANDARD_MENU_PAD_X, STANDARD_MENU_PAD_Y, 0,
      background, selected_background, disabled);
  cairo_surface_destroy (background);
  cairo_surface_destroy (selected_background);
  cairo_surface_destroy (disabled);

  return menu;
}

int
standard_menu_add_tag (Menu *menu, const char *title, int fontsize)
{
  return cairo_menu_add_item_full (menu->menu, NULL, CAIRO_MENU_IMAGE_POSITION_LEFT, title,
      fontsize, CAIRO_MENU_DEFAULT_TEXT_COLOR, CAIRO_MENU_ALIGN_MIDDLE_CENTER,
      menu->menu->default_item_width / 2, menu->menu->default_item_height,
      STANDARD_MENU_ITEM_IPAD_X, STANDARD_MENU_ITEM_IPAD_Y, FALSE,
      NULL, NULL, NULL, NULL);
}

int
standard_menu_add_item (Menu *menu, const char *title, int fontsize)
{
  int idx;

  idx = cairo_menu_add_item (menu->menu, title, fontsize);
  menu->menu->items[idx].ipad_x = STANDARD_MENU_ITEM_IPAD_X;
  menu->menu->items[idx].ipad_y = STANDARD_MENU_ITEM_IPAD_Y;

  return idx;
}

void
free_text (Menu *menu)
{
  if (menu->text.lines != NULL && menu->text.lines[0] != NULL)
    free (menu->text.lines[0]);
  if (menu->text.lines != NULL)
    free (menu->text.lines);
}

cairo_surface_t *
create_standard_gauge (int width, int height, double display_scale,
  unsigned int percent, float dr, float dg, float db, float r, float g, float b) {
  cairo_hires_buffer_t *donebg;
  cairo_hires_buffer_t *bg;
  int done_width;
  cairo_hires_buffer_t *gaugebg;
  cairo_surface_t *gaugesurface;
  cairo_t *cr;
  cairo_pattern_t *linpat = NULL;

  gaugebg = cairo_hires_buffer_create (CAIRO_FORMAT_ARGB32, width, height,
    display_scale);

  if (percent > 100)
    percent = 100;
  done_width = (width - (2 * STANDARD_MENU_BOX_X)) * percent / 100;


  donebg = cairo_menu_create_hires_default_background (done_width,
      height - (2 * STANDARD_MENU_BOX_Y), display_scale, dr, dg, db);
  bg = cairo_menu_create_hires_default_background (width - (2 * STANDARD_MENU_BOX_X),
      height - (2 * STANDARD_MENU_BOX_Y), display_scale, r, g, b);

  // Draw the grey/silver gradient (for the border)
  cr = cairo_hires_buffer_cairo (gaugebg);
  cairo_utils_clip_round_edge (cr, width, height,
      STANDARD_MENU_BOX_CORNER_RADIUS, STANDARD_MENU_BOX_CORNER_RADIUS,
      STANDARD_MENU_BOX_CORNER_RADIUS);

  linpat = cairo_pattern_create_linear (width, 0, width, height);

  cairo_pattern_add_color_stop_rgb (linpat, 0.0, 0.3, 0.3, 0.3);
  cairo_pattern_add_color_stop_rgb (linpat, 1.0, 0.7, 0.7, 0.7);

  cairo_set_source (cr, linpat);
  cairo_paint_with_alpha (cr, 0.5);

  // Clear the inside of the gradient in order to create a border
  cairo_utils_clip_round_edge (cr, width, height,
      STANDARD_MENU_BOX_CORNER_RADIUS + STANDARD_MENU_BOX_BORDER_WIDTH,
      STANDARD_MENU_BOX_CORNER_RADIUS + STANDARD_MENU_BOX_BORDER_WIDTH,
      STANDARD_MENU_BOX_CORNER_RADIUS);

  cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
  cairo_paint (cr);

  // Add a 40% alpha black filter inside, creating a frame around the button
  cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
  cairo_set_source_rgb (cr, 0, 0, 0);
  cairo_paint_with_alpha (cr, 0.4);

  // Draw the actual button
  cairo_hires_buffer_paint (bg, cr, STANDARD_MENU_BOX_X, STANDARD_MENU_BOX_Y);
  cairo_hires_buffer_paint (donebg, cr, STANDARD_MENU_BOX_X, STANDARD_MENU_BOX_Y);
  cairo_pattern_destroy (linpat);
  cairo_hires_buffer_destroy (donebg);
  cairo_hires_buffer_destroy (bg);

  // Return just the surface
  gaugesurface = cairo_surface_reference (cairo_hires_buffer_surface (gaugebg));
  cairo_hires_buffer_destroy (gaugebg);

  return gaugesurface;
}

void standard_menu_update_gauge (Menu *menu, unsigned int percent) {
  CairoMenuItem *item = &menu->menu->items[0];
  char percent_text[6];
  cairo_surface_t *gauge;

  if (percent > 100)
    percent = 100;

  gauge = create_standard_gauge (STANDARD_MENU_ITEM_BOX_WIDTH,
      STANDARD_MENU_ITEM_BOX_HEIGHT,
      cairo_hires_buffer_scale (cairo_menu_get_hires_buffer (menu->menu)),
      percent, menu->gauge_rgb[0], menu->gauge_rgb[1], menu->gauge_rgb[2],
      menu->gauge_rgb[3], menu->gauge_rgb[4], menu->gauge_rgb[5]);
  snprintf (percent_text, sizeof(percent_text), "%d%%", percent);

  //item->enabled = FALSE;
  if (item->text)
    free (item->text);
  item->text = strdup (percent_text);
  cairo_surface_destroy (item->bg_image);
  cairo_surface_destroy (item->bg_sel_image);
  item->bg_image = cairo_surface_reference (gauge);
  item->bg_sel_image = cairo_surface_reference (gauge);
  cairo_surface_destroy (gauge);
}
